import hashlib
import shutil
from enum import Enum
from pathlib import Path
from typing import List, Optional, Set

import git
import httpx
from httpx import Client

from biorepo.exception import SourceException


class SourceEnum(Enum):
    URL = "url"
    GIT = "git"
    LOCAL = "local"


class SourceManager:
    def __init__(self, name: str, root: Path):
        self.name = name
        self.root = root
        self.os = ""

    @property
    def path(self) -> Path:
        h = hashlib.sha256()
        h.update(self.name.encode())
        h = h.hexdigest()
        p = Path(h[0:2], h[2:4], h[4:], self.name)
        return self.root.joinpath(p)

    @property
    def source_path(self) -> Path:
        return self.path / "source"

    @property
    def bin_path(self) -> Path:
        return self.path / "bin"

    @property
    def version_path(self) -> Path:
        return self.path / "version"

    def remove_source(self):
        if self.source_path.exists():
            shutil.rmtree(self.source_path)

    def remove(self):
        if self.path.exists():
            shutil.rmtree(self.path)

    def cache(self):
        if self.path.exists():
            return True
        return False


class BaseSource(SourceManager):
    def __init__(
        self,
        name: str,
        root: Path,
        bin: List[str],
        git_url: Optional[str] = None,
        git_branch: Optional[str] = None,
        git_commit: Optional[str] = None,
        git_recursive: bool = False,
        url: Optional[str] = None,
        user_agent: Optional[str] = None,
        path: Optional[Path] = None,
        **kwargs,
    ):
        super().__init__(name, root)
        self.git_url = str(git_url) if git_url else None
        self.git_branch = git_branch
        self.git_commit = git_commit
        self.git_recursive = git_recursive
        self.download_url = str(url) if url else None
        self.user_agent = user_agent
        self.local_path = path
        if self.download_url:
            self.source_type = SourceEnum.URL
        elif self.git_url:
            self.source_type = SourceEnum.GIT
        elif self.local_path:
            self.source_type = SourceEnum.LOCAL
        else:
            raise Exception("No source provided")
        self.group: Set[str] = set()
        self.deps: List[BaseSource] = []
        self.bin = bin

    def add_dep(self, dep: "BaseSource"):
        self.deps.append(dep)

    def set_group(self, group: str):
        self.group.add(group)

    def create_source(self):
        raise NotImplementedError

    def copy_bin(self):
        if not self.bin_path.exists():
            self.bin_path.mkdir(parents=True, exist_ok=True)
        for b in self.bin:
            exist_bin_path = self.source_path / b
            if not exist_bin_path.exists():
                raise SourceException(f"Bin {b} not found")
            shutil.copy(exist_bin_path, self.bin_path)

    @property
    def source_uri(self):
        if self.__class__ == GitSource:
            return str(self.git_url)
        elif self.__class__ == UrlSource:
            return str(self.download_url)
        else:
            return str(self.local_path)

    @property
    def version(self) -> str:
        if not self.source_path.exists():
            return ""
        if self.version_path.exists():
            with open(self.version_path, "r") as f:
                return f.read()
        return ""

    @version.setter
    def version(self, version: str):
        if not self.version_path.parent.exists():
            self.version_path.parent.mkdir(parents=True, exist_ok=True)
        with open(self.version_path, "w") as f:
            f.write(str(version))


class UrlSource(BaseSource):
    def get_response_modtime(self, url: str) -> float:
        try:
            client = Client(headers={"user-agent": self.user_agent or "biorepo/0.0.1"})
        except:  # noqa: E722
            return 0.0
        r = client.head(url)
        if "Last-Modified" not in r.headers:
            return 0.0
        return float(r.headers["Last-Modified"]) or 0.0

    def create_source(self):
        assert self.download_url
        old_version = float(self.version or 0.0)
        new_version = self.get_response_modtime(self.download_url)
        if new_version < old_version:
            return
        self.version = str(new_version)
        if self.source_path.exists():
            shutil.rmtree(self.source_path)
        self.source_path.mkdir(parents=True, exist_ok=True)
        assert self.download_url
        client = Client()
        try:
            r = client.get(
                self.download_url,
            )
            with open(self.source_path / self.name, "wb") as f:
                f.write(r.content)
        except httpx.HTTPError as e:
            client.close()
            self.remove()
            raise SourceException(f"Download {self.download_url} failed: {e}")
        except Exception as e:
            self.remove()
            raise SourceException(f"Download {self.download_url} failed: {e}")

    def __str__(self):
        return "URL(url=%s)" % self.download_url


class GitSource(BaseSource):
    class GitVersion:
        def __init__(
            self,
            url: str,
            commit: str,
            branch: str,
        ):
            self.url = url
            self.commit = commit
            self.branch = branch

        def __eq__(self, other: "GitSource.GitVersion"):
            return (
                self.url == other.url
                and self.commit == other.commit
                and self.branch == other.branch
            )

        def __str__(self) -> str:
            return f"{self.url}\n{self.commit}\n{self.branch}"

        @classmethod
        def from_str(cls, s: str) -> "GitSource.GitVersion":
            if not s:
                return cls("", "", "")
            url, commit, branch = s.split("\n")
            return cls(url, commit, branch)

        def save(self, path: Path):
            with open(path, "w") as f:
                f.write(str(self))

    def _create_source(self):
        assert self.git_url
        git_version = self.GitVersion(
            url=self.git_url,
            commit=self.git_commit or "",
            branch=self.git_branch or "",
        )

        old_version = self.GitVersion.from_str(self.version)
        if git_version == old_version:
            return

        self.version = str(git_version)
        if self.source_path.exists():
            shutil.rmtree(self.source_path)

        self.path.mkdir(parents=True, exist_ok=True)
        assert self.git_url
        git.Repo.clone_from(
            url=self.git_url,
            to_path=self.source_path,
            branch=self.git_branch,
            recursive=self.git_recursive,
        )
        if self.git_commit:
            repo = git.Repo(self.source_path)
            repo.git.checkout(self.git_commit)
            repo.git.reset("--hard", self.git_commit)

    def create_source(self):
        try:
            self._create_source()
        except git.GitCommandError as e:
            self.remove()
            raise SourceException(f"Git clone {self.git_url} failed: {e}")


class LocalSource(BaseSource):
    @staticmethod
    def get_local_version(path: Path, now: float) -> float:
        paths: List[Path] = []
        paths.append(path)
        max_mtime = 0
        while paths:
            p = paths.pop()
            if p.is_file():
                max_mtime = max(max_mtime, p.stat().st_mtime)
                if now < max_mtime:
                    return max_mtime
            elif p.is_dir():
                for pp in p.iterdir():
                    paths.append(pp)
        return now

    def _create_source(self):
        assert self.local_path
        old_version = float(self.version or 0.0)
        new_version = self.get_local_version(self.local_path, old_version)
        if new_version <= old_version:
            return

        self.version = str(new_version)
        if self.source_path.exists():
            shutil.rmtree(self.source_path)
        shutil.copytree(self.local_path, self.source_path)

    def create_source(self):
        try:
            self._create_source()
        except Exception as e:
            self.remove()
            raise SourceException(f"Copy {self.local_path} failed: {e}")



from typing import Optional, Union, List, Sequence

from rich.console import Console
from rich.progress import Progress, ProgressColumn
from rich.theme import Theme
from rich.table import Table

# copy from pdm/termui.py

DEFAULT_THEME = {
    "primary": "cyan",
    "success": "green",
    "warning": "yellow",
    "error": "red",
    "info": "blue",
    "req": "bold green",
}

_console = Console(highlight=False, theme=Theme(DEFAULT_THEME))


def is_legacy_windows(console: Optional[Console] = None):
    if console is None:
        console = _console
    return console.legacy_windows


class Emoji:
    if is_legacy_windows():
        SUCC = "v"
        FAIL = "x"
        LOCK = " "
        CONGRAT = " "
        POPPER = " "
        ELLIPSIS = "..."
        ARROW_SEPARATOR = ">"
    else:
        SUCC = ":heavy_check_mark:"
        FAIL = ":heavy_multiplication_x:"
        LOCK = ":lock:"
        POPPER = ":party_popper:"
        ELLIPSIS = "…"
        ARROW_SEPARATOR = "➤"


if is_legacy_windows():
    SPINNER = "line"

else:
    SPINNER = "dots"


def info(msg: str, *args, **kwargs):
    _console.print(msg, style="info", *args, **kwargs)


def error(msg: str, *args, **kwargs):
    _console.print(msg, style="error", *args, **kwargs)


def ask(msg: str, default_value: str):
    return _console.input(msg) or default_value

def choice(msg: str, choices: List[str]):
    return _console

class UI:
    def open_spinner(self, text: str):
        return _console.status(text, spinner=SPINNER, spinner_style="primary")

    def make_progress(self, *columns: Union[str, ProgressColumn], **kwargs) -> Progress:
        return Progress(
            *columns,
            console=_console,
            **kwargs,
        )

    def table(self, raws: List[Sequence[str]], header: List[str]):
        table = Table(
            *header,
        )
        for raw in raws:
            raw = [str(i) for i in raw]
            table.add_row(*raw)
        return table



import os
from enum import Enum
from pathlib import Path
from subprocess import PIPE, Popen
from typing import List, Optional

from rich.progress import Progress

from .env import Envirment
from .exception import ShellException


class ShellEnum(Enum):
    PYTHON = "python"
    CMD = "shell"
    OTHER = 3

    def __str__(self):
        return self.name.lower()


class Shell:
    def __init__(
        self,
        source_path: Path,
        download_path: Path,
        run: List[str],
        envs: Optional[List[Envirment]] = None,
        shell_type: ShellEnum = ShellEnum.CMD,
    ):
        self.source_path = source_path.absolute()
        self.download_path = download_path.absolute()
        self.shell_type = shell_type
        self.envs = envs or []
        self.run = run or []

    def execute(self, progress: Progress):
        raise NotImplementedError

    def __str__(self):
        return f"{self.shell_type} {self.run}"

    def __repr__(self) -> str:
        return "Shell(working_dir={self.working_dir}, shell_type={self.shell_type}, envs={self.envs}, run={self.run})"


class CMDShell(Shell):
    def set_env(self):
        for env in self.envs:
            env.set()

    def unset_env(self):
        for env in self.envs:
            env.unset()

    def replace_cmd(self, cmd: str):
        if "{PATH}" in cmd:
            return cmd.replace("{PATH}", str(self.download_path))
        if "{SOURCE_PATH}" in cmd:
            return cmd.replace("{SOURCE_PATH}", str(self.source_path))
        return cmd

    def _execute(self, run: str):
        self.set_env()
        cmd = Popen(
            run,
            stdout=PIPE,
            stderr=PIPE,
            shell=True,
            cwd=self.source_path,
            env=os.environ.copy(),
        ).communicate()
        stdout, stderr = cmd
        self.unset_env()
        return stdout, stderr

    def execute(self, progress: Progress):
        if not self.run:
            return

        for run in self.run:
            run = self.replace_cmd(run)
            stderr = ""
            try:
                stdout, stderr = self._execute(run)
                # TODO: handle stderr
                stderr = ""
            except Exception as e:
                stderr = str(e)
            if stderr:
                raise ShellException(stderr)


class PythonShell(Shell):
    def execute(self):
        ...



import enum
import multiprocessing
from concurrent.futures import ThreadPoolExecutor
from typing import Dict, List, Union, Optional

from rich.progress import Progress, SpinnerColumn, TaskProgressColumn

from biorepo import ui
from biorepo.exception import BioReopException
from biorepo.shell import Shell
from biorepo.source import BaseSource, SourceEnum


class RunStatus(enum.Enum):
    SUCCESS = 1
    FAIL = 2
    RUNNING = 3


class Run:
    def __init__(self, shell: Shell, source: BaseSource):
        self.status = RunStatus.RUNNING
        self.shell = shell
        self.source = source

    def install(self, progress: Progress):
        job = progress.add_task(
            f"Installing... [req]{self.source.name}[/]", text="", total=None
        )
        try:
            self.source.create_source()
            self.shell.execute(progress)
            self.source.copy_bin()            
            self.status = RunStatus.SUCCESS
            progress.live.console.print(
                f"  [success]{ui.Emoji.SUCC}[/] Install [req]{self.source.name}[/] successful"
            )
        except BioReopException as e:
            self.status = RunStatus.FAIL
            progress.live.console.print(
                f"  [error]{ui.Emoji.FAIL}[/] Install [primary]{self.source.name}[/] failed"
            )
            progress.live.console.print(f"    [error]{e.msg}[/]")
        finally:
            progress.update(job, visible=False)

    def remove(self, progress: Progress):
        job = progress.add_task(
            f"Removing... [req]{self.source.name}[/]", text="", total=None
        )
        try:
            self.source.remove()
            self.status = RunStatus.SUCCESS
            progress.live.console.print(
                f"  [success]{ui.Emoji.SUCC}[/] Remove [req]{self.source.name}[/] successful"
            )
        except BioReopException as e:
            self.status = RunStatus.FAIL
            progress.live.console.print(
                f"  [error]{ui.Emoji.FAIL}[/] Remove [primary]{self.source.name}[/] failed"
            )
            progress.live.console.print(f"    [error]{e.msg}[/]")
        finally:
            progress.update(job, visible=False)


class Install:
    def __init__(self, nthread: int = 4):
        self.runs: Dict[str, Run] = {}
        self.nthread = nthread

    def create_executor(self):
        return ThreadPoolExecutor(
            min(len(self.runs), self.nthread, multiprocessing.cpu_count()) or 1
        )

    def add_run(self, r: Union[Run, List[Run]]):
        if not isinstance(r, list):
            r = [r]
        for run in r:
            self.runs[run.source.name] = run

    def executor(self, runs: List[Run], action: str):
        pool = self.create_executor()
        with ui.UI().make_progress(
            " ",
            SpinnerColumn(ui.SPINNER, speed=1, style="primary"),
            "{task.description}",
            "[info]{task.fields[text]}",
            TaskProgressColumn("[info]{task.percentage:>3.0f}%[/]"),
        ) as progress:
            live = progress.live
            live.console.print("")
            for run in runs:
                pool.submit(getattr(run, action), progress)
            pool.shutdown(wait=True)

        failed = [
            run for name, run in self.runs.items() if run.status == RunStatus.FAIL
        ]
        if not failed:
            live.console.print(f"{ui.Emoji.POPPER} All complete!")

    def install(self):
        if not self.runs:
            ui.error("No source to install")
            exit(0)
        self.executor(list(self.runs.values()), "install")

    def remove(self, names: Optional[List[str]] = None):
        runs = []
        if names:
            for name in names:
                if name in self.runs:
                    runs.append(self.runs[name])
                else:
                    ui.error(f"Source [primary]{name}[/] not found")
                    exit(0)
        else:
            runs = list(self.runs.values())
        if not runs:
            exit(0)
        self.executor(runs, "remove")

    def list(self):
        headers = ["name", "source"]
        raws = []
        for name, run in self.runs.items():
            if run.source.source_type == SourceEnum.GIT:
                name = f"[bold yellow]{name}[/]"
            elif run.source.source_type == SourceEnum.URL:
                name = f"[bold green]{name}[/]"
            else:
                name = f"[bold blue]{name}[/]"
            uri = f"[link={run.source.source_uri} cyan]{run.source.source_uri}[/]"
            raws.append([name, uri])
        table = ui.UI().table(raws=raws, header=headers)
        ui._console.print(table)



import os
from typing import Any, Dict, List, Optional


class Envirment:
    def __init__(self, name: str, value: Optional[str] = None):
        self._value = value is None
        self.name = name
        self.value = value

    def lower(self):
        if self.value is None:
            return None
        return self.name.lower()

    def upper(self):
        if self.value is None:
            return None
        return self.name.upper()

    def to_dict(self):
        return {"name": self.name, "value": self.value}

    def __repr__(self):
        return f"<Envirment name={self.name} value={self.value}>"

    def __str__(self):
        return f"{self.name}={self.value}"

    def __eq__(self, other):
        if isinstance(other, Envirment):
            return self.name == other.name and self.value == other.value
        return False

    @classmethod
    def from_environ(cls) -> List["Envirment"]:
        return [Envirment(name, os.environ.get(name)) for name in os.environ]

    @classmethod
    def from_dict(cls, dict: Dict[str, Any]) -> List["Envirment"]:
        return [
            Envirment(env_name, str(env_value)) for env_name, env_value in dict.items()
        ]

    @classmethod
    def from_list(cls, envs: List[str]) -> List["Envirment"]:
        return [Envirment(env) for env in envs]

    def set(self):
        os.environ[self.name] = self.value or ""

    def unset(self):
        if self.name in os.environ:
            del os.environ[self.name]



from typing import Optional


class BioReopException(Exception):
    """Base class for all exceptions raised by the biorepo package."""

    def __init__(self, msg: Optional[str] = None):
        self.msg = msg

    def __str__(self):
        return f"{self.__class__.__name__}: {self.msg}"


class BioRepoMissItem(BioReopException):
    def __init__(self, msg: str, key: str):
        self.key = key
        super().__init__("%s %s" % (msg, key))


class SourceException(BioReopException):
    """Base class for all exceptions raised by the source package."""

    def __init__(self, msg: Optional[str] = None):
        super().__init__(msg)


class ShellException(BioReopException):
    """Base class for all exceptions raised by the shell package."""

    def __init__(self, msg: Optional[str] = None):
        super().__init__(msg)


class InstallException(BioReopException):
    """Base class for all exceptions raised by the install package."""

    def __init__(self, msg: Optional[str] = None):
        super().__init__(msg)



import shutil
import json
import sys
from pathlib import Path
from typing import Dict, Union, Optional, List

import subprocess

from biorepo.biorepo import BioRepo, GitSourceModel, LoaclSourceModel
from biorepo.env import Envirment
from biorepo.install import Install, Run
from biorepo.shell import CMDShell, PythonShell, Shell, ShellEnum
from biorepo.source import GitSource, LocalSource, SourceEnum, UrlSource
from biorepo import ui


class Repo(object):
    def __init__(
        self,
        root: Path,
        biorepo_path: Path,
        nthread: int = 4,
    ):
        self.root = root
        self.dwonload_path = self.root / "download"
        self.mainfest_path = self.root / "mainfest.json"
        self.bin_path = self.root / "bin"
        self._install = Install(nthread=nthread)
        self.biorepo = BioRepo().load(str(biorepo_path))
        self.biorepo_path = biorepo_path
        assert self.biorepo
        self.sources: Dict[str, Union[GitSource, LocalSource, UrlSource]] = {}
        self.bins: Dict[str, Path] = {}
        self._create_runs()

    def mainfest(self):
        with open(self.mainfest_path, "r", encoding="utf-8") as f:
            data = f.read()
        return json.loads(data)

    def set_mainfest(self, key: str, value):
        if not self.mainfest_path.exists():
            self.mainfest_path.touch()
            self.mainfest_path.write_text("{}")

        with open(self.mainfest_path, "r", encoding="utf-8") as f:
            raw_data = json.loads(f.read())
        if "." in key:
            keys = key.split(".")
        else:
            keys = [key]
        data = raw_data
        for key in keys[:-1]:
            if key not in data:
                data[key] = {}
            data = data[key]
        data[keys[-1]] = value
        with open(self.mainfest_path, "w", encoding="utf-8") as f:
            f.write(json.dumps(raw_data, indent=4))

    def _create_runs(self):
        assert self.biorepo
        source_handle = {
            SourceEnum.GIT: GitSource,
            SourceEnum.LOCAL: LocalSource,
            SourceEnum.URL: UrlSource,
        }

        shell_handle = {
            ShellEnum.CMD: CMDShell,
            ShellEnum.PYTHON: PythonShell,
            ShellEnum.OTHER: Shell,
        }

        sources = self.biorepo.get_source_list()
        for name, source in sources.items():
            source_dict = source.model_dump()
            source_dict["root"] = self.dwonload_path
            if isinstance(source, GitSourceModel):
                source_type = SourceEnum.GIT
            elif isinstance(source, LoaclSourceModel):
                source_type = SourceEnum.LOCAL
            else:
                source_type = SourceEnum.URL
            shell_type = source.shell or ShellEnum.CMD

            run_command = source_dict.pop("run")
            run_envs = Envirment.from_dict(source_dict.pop("envs"))
            source = source_handle[source_type](**source_dict)
            shell = shell_handle[shell_type](
                source_path=source.source_path,
                download_path=source.path,
                run=run_command,
                envs=run_envs,
            )

            self.sources[name] = source
            run = Run(shell=shell, source=source)
            self._install.add_run(run)

    def list(self):
        self._install.list()

    def install(self):
        self._install.install()
        if not self.bin_path.exists():
            self.bin_path.mkdir(parents=True, exist_ok=True)
        for name, source in self.sources.items():
            for b in source.bin:
                exist_bin_path = source.source_path / b
                if not exist_bin_path.exists():
                    ui.error(f"bin {b} not found in {name}")
                shutil.copy(exist_bin_path, self.bin_path)
                self.set_mainfest(f"bin.{b}", str(self.bin_path / b))

    def remove(self, names: Optional[List[str]] = None):
        if not names:
            names = list(set(self.biorepo.sources.sources_names))
        self._install.remove(names)
        for name in names:
            self.biorepo.remove_source(name)
        self.biorepo.dump(str(self.biorepo_path))

    def run(self, name: str, *argv):
        bins = self.mainfest().get("bin", {})
        if name not in bins:
            ui.error(f"source {name} not found")
            return
        exe = bins[name]
        subprocess.Popen(
            executable=exe,
            args=argv,
        ).communicate()

    def check_require(self):
        assert self.biorepo
        require = self.biorepo.get_require()
        if not require:
            return
        spinner = ui.UI().open_spinner("Checking requirements")
        spinner.start()
        for r in require:
            try:
                spinner.update(f"Checking requirement [bold blue]{r}[/]")
                stdout, stderr = subprocess.Popen(
                    f"{r}",
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    encoding="utf-8",
                    shell=True,
                ).communicate()
                if "not found" in stderr or "not found" in stdout:
                    spinner.stop()
                    ui.error(f"Requirement [bold blue]{r}[/] not found")
                    exit(0)
            except Exception as e:
                spinner.stop()
                ui.error(f"Requirement [bold blue]{r}[/] not found")
                ui.error(str(e))
                exit(0)
        spinner.stop()



import os
import sys
from pathlib import Path

from typing import List, Optional
from typer import Option, Typer, Argument

from biorepo.repo import Repo
from biorepo.biorepo import BioRepo, OSEnum
from biorepo.ui import error, info, ask

app = Typer(
    name="biorepo",
    help="A tool to manage bioinformatics software",
    add_completion=False,
    pretty_exceptions_enable=False,
)


@app.command()
def install(
    file: Path = Option(Path("biorepo.toml")),
    prefix: Path = Option(Path(".biorepo")),
):
    if not file.exists():
        error(f"biorepo.toml not found in {file}")
        info("Please run `biorepo new` to create a biorepo.toml")
        return
    repo = Repo(root=prefix, biorepo_path=file)
    repo.check_require()
    repo.install()


@app.command()
def remove(
    file: Path = Option(Path("biorepo.toml")),
    prefix: Path = Option(Path(".biorepo")),
    names: Optional[List[str]] = Argument(None),
):
    if not file.exists():
        error(f"biorepo.toml not found in {file}")
        return
    repo = Repo(root=prefix, biorepo_path=file)
    repo.remove(names)


@app.command()
def list(
    file: Path = Option(Path("biorepo.toml")),
    prefix: Path = Option(Path(".biorepo")),
):
    if not file.exists():
        error(f"biorepo.toml not found in {file}")
        return
    repo = Repo(
        root=prefix,
        biorepo_path=file,
    )
    assert repo
    repo.list()


@app.command()
def new(
    file: Path = Option(Path("biorepo.toml")),
):
    p = Path(os.path.curdir).cwd()
    biorepo_file = p / file
    dirname = p.name
    if biorepo_file.exists():
        error(f"{file} already exists")
        goon = ask("Continue? [[cyan]y/n[/]]: ", "n")
        if goon.lower() != "y":
            return
    name = ask(f"Biorepo name [[cyan]{dirname}[/]]: ", dirname)
    version = ask("Version [[cyan]0.0.1[/]]: ", "0.0.1")
    description = ask("Description: ", " ")
    bio = BioRepo.new(
        name=name,
        version=version,
        description=description,
        os=[OSEnum.linux],
    )
    if not file.parent.exists():
        file.parent.mkdir(parents=True, exist_ok=True)
    bio.dump(str(biorepo_file))


def run(
    name: str,
    argv: List[str],
):
    file = Path("biorepo.toml")
    prefix = Path(".biorepo")
    if not file.exists():
        error(f"biorepo.toml not found in {file}")
        return
    if not prefix.exists():
        error("This is not a biorepo directory")
        return
    repo = Repo(
        root=prefix,
        biorepo_path=file,
    )
    assert repo
    print(*sys.argv[sys.argv.index(name) + 1 :])
    repo.run(name, *sys.argv[sys.argv.index(name) + 1 :])

@app.command(name='run')
def typer_run(
    name: str,
    argv: List[str],
):
    pass


def main():
    if len(sys.argv) > 1 and sys.argv[1] == "run":
        if len(sys.argv) >= 3:
            run(sys.argv[2], sys.argv[3:])
            return
    app()



import datetime
import enum
import sys
from collections import OrderedDict
from typing import Dict, List, Optional, Union

import chardet
import pydantic
import tomlkit
from pydantic import BaseModel

from biorepo import exception
from biorepo.shell import ShellEnum
from biorepo.source import SourceEnum
from biorepo.ui import _console


msg_replace = {
    "too_short": "Input is empty",
    "missing": "%s is required",
    "list_type": "%s must be a list",
}


def rich_error(title: str, e: pydantic.ValidationError, data: Dict):
    errors = e.errors()
    key_errors = {}
    for err in errors:
        for key in err["loc"]:
            if not isinstance(key, str):
                continue
            if key not in data:
                data[key] = ""
            if key not in key_errors:
                key_errors[key] = {}
            err["msg"] = msg_replace.get(err["type"], err["msg"])
            if "%" in err["msg"]:
                err["msg"] = err["msg"] % key
            if err["type"] == "list_type":
                err["msg"] = (
                    err["msg"]
                    + "  [blue]fix: [green][[/] %s [green]][/][/]" % data[key]
                )

            key_errors[key]["msg"] = err["msg"]
    _console.print(f"{title}")
    for key, value in data.items():
        if key in key_errors:
            key_len = len(key) + 3
            space = " " * key_len
            _console.print("")
            _console.print(f"[red bold]{key} = {value}[/]")
            _console.print("[red bold]%s↑" % space)
            _console.print(f"[red bold]%sERROR: {key_errors[key]['msg']}[/]" % space)
            _console.print("")
        else:
            _console.print(f"{key} = {value}")
    exit(1)


class OSEnum(enum.Enum):
    windows = "windows"
    linux = "linux"
    mac = "macos"

    def __str__(self):
        return self.value


class Base(BaseModel):
    ...


class BioRepoMeta(Base):
    version: str = pydantic.Field(min_length=1)
    name: str = pydantic.Field(min_length=1)
    os: List[OSEnum]
    description: Optional[str] = None
    author: Optional[str] = None
    email: Optional[pydantic.EmailStr] = None
    home: Optional[pydantic.HttpUrl] = None
    date: Optional[datetime.date] = None
    require: Optional[List[str]] = None


class SourceModel(Base):
    name: str = pydantic.Field(min_length=1)
    bin: List[str]
    shell: Optional[ShellEnum] = ShellEnum.CMD
    group: Optional[Union[str, List[str]]] = None
    run: Optional[List[str]] = None
    envs: Optional[Dict[str, str]] = None
    remove: Optional[bool] = False


class URLSourceModel(SourceModel):
    url: pydantic.HttpUrl
    user_agent: Optional[str] = None
    proxy: Optional[pydantic.HttpUrl] = None
    username: Optional[str] = None
    password: Optional[str] = None
    token: Optional[str] = None


class LoaclSourceModel(SourceModel):
    path: Union[pydantic.FilePath, pydantic.DirectoryPath]


class GitSourceModel(SourceModel):
    git_url: pydantic.AnyUrl
    git_branch: Optional[str] = None
    git_commit: Optional[str] = None
    git_recursive: bool = False
    git_mirror: Optional[str] = None


source_map = {
    SourceEnum.URL: URLSourceModel,
    SourceEnum.LOCAL: LoaclSourceModel,
    SourceEnum.GIT: GitSourceModel,
}

ENV_COMMON = "__ENV_COMMON__"


class EnvManager:
    def __init__(self):
        self.data: Dict[OSEnum, Dict[str, Dict[str, str]]] = {
            OSEnum.linux: {ENV_COMMON: {}},
            OSEnum.windows: {ENV_COMMON: {}},
            OSEnum.mac: {ENV_COMMON: {}},
        }

    def set(
        self,
        key: str,
        value: str,
        target_bin: Optional[str] = None,
        target_os: Optional[OSEnum] = None,
    ):
        if target_bin is None:
            target_bin = ENV_COMMON
        update_os: List[OSEnum] = []
        if target_os is None:
            update_os.extend(
                [
                    OSEnum.linux,
                    OSEnum.windows,
                    OSEnum.mac,
                ]
            )
        else:
            update_os.append(target_os)
        for os in update_os:
            if target_bin not in self.data[os]:
                self.data[os][target_bin] = {}
            self.data[os][target_bin][key] = str(value)

    def get(
        self,
        key: str,
        default: Optional[str] = None,
        target_bin: Optional[str] = None,
        target_os: Optional[OSEnum] = None,
    ):
        if target_bin is None:
            target_bin = ENV_COMMON
        if target_os is None:
            target_os = OSEnum.linux
        if target_bin not in self.data[target_os]:
            return None
        return (
            self.data[target_os][target_bin].get(key)
            or self.data[target_os][ENV_COMMON].get(key)
            or default
        )

    def get_bin_envs(self, target_bin: str, target_os: OSEnum) -> Dict[str, str]:
        envs = {}
        common = self.data[target_os][ENV_COMMON]
        envs.update(common)
        if target_bin in self.data[target_os]:
            bin_env = self.data[target_os][target_bin]
            envs.update(bin_env)
        return envs


class SourceManager:
    def __init__(self, env: EnvManager):
        self.sources: Dict[
            OSEnum, Dict[str, Union[URLSourceModel, LoaclSourceModel, GitSourceModel]]
        ] = {
            OSEnum.linux: {},
            OSEnum.windows: {},
            OSEnum.mac: {},
        }
        self._env = env
        self.sources_names: List[str] = []
        self.sources_raw: Dict[str, Dict] = {}

    def add_raw(self, flag: str, source: Dict):
        self.sources_raw[flag] = source

    def add_source(
        self,
        source: Union[LoaclSourceModel, URLSourceModel, GitSourceModel],
        target_os: Optional[OSEnum],
    ):
        update_os = []
        if target_os is None:
            update_os = [OSEnum.linux, OSEnum.mac, OSEnum.windows]
        else:
            update_os = [target_os]
        for os in update_os:
            self.sources[os][source.name] = source
            if not source.envs:
                source.envs = {}
            source.envs.update(self._env.get_bin_envs(source.name, os))
        self.sources_names.append(source.name)

    def del_source(self, name: str):
        for os in list(OSEnum):
            if f"name.{os}" in self.sources_raw:
                self.sources_raw[f"name.{os}"]["remove"] = True
        if name in self.sources_raw:
            self.sources_raw[name]["remove"] = True

    def list(self):
        sys_os = get_os()
        sources = self.sources[sys_os]
        return sources


def get_os() -> OSEnum:
    platform = sys.platform
    if platform == OSEnum.linux.value:
        return OSEnum.linux
    elif platform == OSEnum.windows.value:
        return OSEnum.windows
    else:
        return OSEnum.mac


class BioRepo:
    def __init__(self):
        self.meta: Optional[BioRepoMeta] = None
        self.envs = EnvManager()
        self.sources = SourceManager(self.envs)

    def get_source_list(self):
        assert self.meta
        if self.meta.os and get_os() not in self.meta.os:
            return {}
        return self.sources.list()

    def get_require(self):
        assert self.meta
        return self.meta.require or []

    @staticmethod
    def detect_encoding(path: str) -> str:
        encoding = "utf-8"
        with open(path, "rb") as fp:
            data = fp.read()
        c = chardet.detect(data).get("encoding") or "utf-8"
        if c != "ascii":
            encoding = c
        return encoding

    @staticmethod
    def guess_source_type(source: Dict):
        if "url" in source:
            return SourceEnum.URL
        elif "git_url" in source:
            return SourceEnum.GIT
        elif "path" in source:
            return SourceEnum.LOCAL
        raise exception.BioReopException(f"Invalid install config {source}")

    def load(self, path: str):
        encoding = self.detect_encoding(path)
        with open(path, "r", encoding=encoding) as fp:
            data = tomlkit.loads(fp.read())
            data = data.value

        if "biorepo" not in data:
            _console.print("[red bold]EOORO:[/] Invalid file: %s" % path)
            exit(1)

        biorepo = data["biorepo"]
        try:
            self.meta = BioRepoMeta(**biorepo)
        except pydantic.ValidationError as e:
            title = "\[biorepo]"  # type: ignore
            rich_error(title, e, biorepo)
        if "env" in data:
            envs = data["env"]
            for key, value in envs.items():
                if not isinstance(value, dict):
                    self.envs.set(key, value)
                    continue
                bin_name = key
                for key, value in value.items():
                    if not isinstance(value, dict):
                        self.envs.set(key, value, target_bin=bin_name)
                        continue
                    if key not in [os.value for os in list(OSEnum)]:
                        _console.print(
                            f"[red bold]ERROR:[/] env {key} must be in %s"
                            % ",".join([i.value for i in list(OSEnum)])
                        )
                        exit(1)
                    os = OSEnum(key)
                    for key, value in value.items():
                        self.envs.set(key, value, target_bin=bin_name, target_os=os)
        if "install" not in data:
            return self
        install = data["install"]
        for bin_name, source in install.items():
            if not isinstance(source, dict):
                raise exception.BioReopException("Invalid %s." % path)
            os_spec = OrderedDict()
            if "linux" in source:
                os_spec[OSEnum.linux] = (source.pop("linux"), "linux")
            if "windows" in source:
                os_spec[OSEnum.windows] = (source.pop("windows"), "windows")
            if "mac" in source:
                os_spec[OSEnum.mac] = (source.pop("mac"), "mac")

            if source:
                for os in list(OSEnum):
                    if os in os_spec:
                        continue
                    os_spec[os] = (source, "")

            oss = list(os_spec)
            oss.reverse()

            for os in oss:
                source = os_spec[os]
                source, display_os = source
                source["name"] = bin_name
                if display_os:
                    self.sources.add_raw(bin_name + f".{display_os}", source)
                else:
                    self.sources.add_raw(bin_name, source)
                source_type = self.guess_source_type(source)
                try:
                    s = source_map[source_type](**source)
                    if s.remove:
                        continue
                    self.sources.add_source(s, target_os=os)
                except pydantic.ValidationError as e:
                    source.pop("name")
                    if display_os:
                        title = f"\[install.{bin_name}.{display_os}]"  # type: ignore
                        rich_error(title, e, source)
                    else:
                        title = f"\[install.{bin_name}]"  # type: ignore
                        rich_error(title, e, source)
        return self

    @classmethod
    def new(
        cls,
        name: str,
        version: str,
        email: Optional[str] = None,
        author: Optional[str] = None,
        description: Optional[str] = None,
        requirement: Optional[List[str]] = None,
        os: Optional[List[OSEnum]] = None,
    ):
        biorepo = BioRepo()
        biorepo.meta = BioRepoMeta(
            name=name,
            version=version,
            description=description,
            require=requirement,
            email=email,
            author=author,
            os=os or [],
        )
        return biorepo

    def remove_source(self, name: str):
        self.sources.del_source(name)

    def dump(self, path: str):
        # write meta
        file = [
            "[biorepo]",
        ]
        assert self.meta
        for key, value in self.meta.model_dump().items():
            if not value:
                # file.append(f'# {key} = ""')
                continue
            if isinstance(value, list):
                v = "[\n  "
                v += ",\n  ".join(['"' + str(i) + '"' for i in value])
                v += "\n]"
                file.append(f"{key} = {v}")
            else:
                v = value
                file.append(f'{key} = "{v}"')
        file.append("")

        file.append("[env]")
        # write env
        if self.envs:
            for os in list(OSEnum):
                for key, value in self.envs.data[os][ENV_COMMON].items():
                    file.append(f'{key} = "{value}"')

            for source_name in self.sources.sources_names:
                for os in list(OSEnum):
                    if source_name not in self.envs.data[os]:
                        continue
                    file.append(f"[env.{source_name}.{os.value}]")
                    for key, value in self.envs.data[os][source_name].items():
                        file.append(f'{key} = "{value}"')
                    file.append("")


        # write install
        sources = self.sources.sources_raw
        if not sources:
            file.append("")
            file.append("[install]")

        for name, source in sources.items():
            source.pop("name")
            file.append(f"[install.{name}]")
            for key, value in source.items():
                if key == "remove":
                    value = "true" if value else "false"
                    file.append("%s = %s" % (key, value))
                    continue
                if isinstance(value, list):
                    file.append(f"{key} = [  ")
                    for item in value:
                        file.append(f'  "{item}",')
                    file.append("]")
                else:
                    v = value
                    file.append(f'{key} = "{v}"')
            file.append("")

        with open(path, "w") as fp:
            fp.write("\n".join(file))
